"""
 Copyright (c) 2023 Anthony Mugendi
 
 This software is released under the MIT License.
 https://opensource.org/licenses/MIT
"""

from os import path

from .types import ensure_type


def path_truncate(path_str: str, max_len: int = 50) -> str:
    """Truncate a path to a given length .

    Args:
        path_str (str): Path to truncate
        max_len (int, optional): Maximum length (chars) to truncate to. Defaults to 50.

    Returns:
        str: Truncated path
    """

    # Validate arguments
    ensure_type(path_str, str, "path_str")
    ensure_type(max_len, int)

    if len(path_str) < max_len:
        return path_str

    path_arr = path_str.split(path.sep)
    arr_e = []
    arr_s = []

    arr_e.append(path_arr[-1])
    char_len = len(path_arr[-1])

    for t in path_arr[0:-1]:
        if char_len < max_len:
            arr_s.append(t)
            char_len += len(t) + 1
        else:
            break

    arr = arr_s + ["..."] + arr_e

    return path.sep.join(arr)



"""
 Copyright (c) 2023 Anthony Mugendi
 
 This software is released under the MIT License.
 https://opensource.org/licenses/MIT
"""

from typing import Union

from .classes import is_class, is_initialized
from .lists import listify


# pick values from list/dict/tuple/class or set
def pick(obj, keys: list, sort_keys: bool = True) -> Union[dict, list, tuple, set]:
    """Returns a subset of items from a class, dict, list, tuple or set\
          based on their keys/props.

    Args:
        obj (class, dict, list, tuple, set): value to pick items from
        keys (list): keys/props to pick
        sort_keys (bool, optional): sort picked keys/props before returning?.\
              Defaults to True.

    Raises:
        TypeError: if obj is not one of class, dict, list, tuple or set

    Returns:
        (dict, list, tuple, set): items picked from class, dict, list, tuple or set
    """
    if not isinstance(obj, (dict, list, tuple, set)) and not is_initialized(obj):
        raise TypeError(
            "obj must be a dict, list, tuple, set or an initialized class object"
        )
    if not isinstance(keys, (tuple, set, list, str, int)):
        raise TypeError("keys to pick must be a string, integer, set, tuple or list")

    keys = listify(keys)

    if sort_keys:
        keys = sorted(keys)

    if isinstance(obj, dict):
        return {k: obj[k] for k in keys if k in obj}

    elif isinstance(obj, tuple):
        length = len(obj)
        return tuple([obj[k] for k in keys if k <= length - 1])

    elif isinstance(obj, set):
        return {k for k in keys if k in obj}

    elif is_class(obj):
        return {k: getattr(obj, k) for k in keys if hasattr(obj, str(k))}

    else:
        length = len(obj)
        return [obj[k] for k in keys if k <= length - 1]



"""
 Copyright (c) 2023 Anthony Mugendi
 
 This software is released under the MIT License.
 https://opensource.org/licenses/MIT
"""

from .classes import *
from .lists import *
from .misc import *
from .numbers import *
from .paths import *
from .system import *
from .types import *



"""
 Copyright (c) 2023 Anthony Mugendi
 
 This software is released under the MIT License.
 https://opensource.org/licenses/MIT
"""

from typing import Any, Union

import numpy as np


def listify(val: Any) -> list:
    """Convert a value into a list.

    Args:
        v (Any): your value

    Returns:
        list: list either containing your input value or value if it was already a list
    """
    if isinstance(val, (tuple, set)):
        val = list(val)
    else:
        val = val if isinstance(val, list) else [val]

    return val


def __flatten(val: Union[set, list, tuple, np.ndarray]) -> iter:
    """Flatten a nested list/set/tuple/numpy array into a single iterable .

    Args:
        val (set, list, tuple, np.ndarray)

    Raises:
        TypeError: if val is one of list/set/tuple/numpy array

    Returns:
        iter: iterabable

    Yields:
        Iterator[iter]: iterabable
    """
    if not isinstance(val, (set, list, tuple, np.ndarray)):
        raise TypeError("Function expects a list, set, tuple or numpy array")

    for i in val:
        if isinstance(i, (list, tuple, np.ndarray)):
            for j in __flatten(i):
                yield j
        else:
            yield i


def flatten(
    val: Union[set, list, tuple, np.ndarray]
) -> Union[set, list, tuple, np.ndarray]:
    """Flattens a list/set/tuple/numpy array to a 1D list/set/tuple/numpy array .

    Args:
        val (set, list, tuple, np.ndarray): your value

    Raises:
        TypeError: if val is one of list/set/tuple/numpy array

    Returns:
        Union[set, list, tuple, np.ndarray]: [description]
    """
    if not isinstance(val, (set, list, tuple, np.ndarray)):
        raise TypeError("Function expects a list, set, tuple or numpy array")

    # run __flatten
    cont = list(__flatten(val=val))

    # return same value as input
    if isinstance(val, np.ndarray):
        return np.array(cont)
    elif isinstance(val, tuple):
        return tuple(cont)
    elif isinstance(val, set):
        return set(cont)
    else:
        return cont


def chunk_arr(arr: list, n: int = 4) -> list:
    """Chunk an array into n - sized chunks .

    Args:
        arr (list): list/array to chunk
        n (int, optional): length of each chunk. Defaults to 4.

    Returns:
        list: chunked array
    """

    # validate inputs
    if not isinstance(arr, list):
        raise TypeError("Expects list, got", arr)
    if not isinstance(n, int):
        raise TypeError("Expects integer, got", n)

    return [arr[i * n : (i + 1) * n] for i in range((len(arr) + n - 1) // n)]



"""
 Copyright (c) 2023 Anthony Mugendi
 
 This software is released under the MIT License.
 https://opensource.org/licenses/MIT
"""


def human_format(value) -> str:
    """Return a human - readable representation of a number .

    Args:
        value ([int,float]): Value you wish to format

    Raises:
        Exception: TypeError

    Returns:
        str: A string representing the number in easy human readable form e.g "2.345K"
    """
    if not isinstance(value, (int, float)):
        raise Exception("Integer/Float value expected")

    num = value
    magnitude = 0
    while abs(num) >= 1000:
        magnitude += 1
        num /= 1000.0

    result = round(value / (1000**magnitude), 3)

    return "{}{}".format(result, ["", "K", "M", "B", "T"][magnitude])



"""
 Copyright (c) 2023 Anthony Mugendi
 
 This software is released under the MIT License.
 https://opensource.org/licenses/MIT
"""

import os
import sys
from typing import Union


def trace_caller(return_trace: bool = False) -> Union[None, tuple]:
    """Log the method that made the call

    Args:
        return_trace (bool, optional): if true, the trace details\
              are returned as a tuple. Defaults to False.

    Returns:
        (None, tuple): prints the trace details or outputs the same as a tuple
    """
    try:
        raise Exception
    except Exception:
        frame = sys.exc_info()[2].tb_frame.f_back.f_back

        trace_str = (
            " >> invoked by:"
            f" {frame.f_code.co_name}\n\tFile:'{frame.f_code.co_filename}\n\tLine:"
            f" {frame.f_code.co_firstlineno}"
        )

        if return_trace:
            return (trace_str, frame.f_code.co_name)

        print(trace_str)


def importing_script():
    # get working directory
    dir = os.getcwd()
    # get script path as passed to python
    script = sys.argv[0]
    # if path is not absolute, then join with current dir
    importing_script = os.path.join(dir, script) if not os.path.isabs(script) else script

    return importing_script



"""
 Copyright (c) 2023 Anthony Mugendi
 
 This software is released under the MIT License.
 https://opensource.org/licenses/MIT
"""

import re
from typing import Any

from .lists import listify

types_obj = {
    "int": "integer",
    "float": "float",
    "complex": "complex",
    "str": "string",
    "list": "list",
    "tuple": "tuple",
    "range": "range",
    "bytes": "bytes",
    "bytearray": "bytearray",
    "memoryview": "memoryview",
    "dict": "dict",
    "bool": "boolean",
    "set": "set",
    "frozenset": "frozenset",
    "NoneType": "NoneType",
}


def get_type_name(_type) -> str:
    """Return the type name of a Python object. \
        This returns full names for types so that str=>"string" and so on...

    Args:
        _type (type): the type you want to get the name of. \
            Example get_type_name(str) -> "string"

    Returns:
        str: full name of the type
    """
    groups = re.search(r"<class\s+'([^']+)'", str(_type)).groups()

    return types_obj[groups[0]]


def ensure_type(val: Any, _types, val_name: str = None) -> None:
    """Ensure that value is of a valid type .

    Args:
        val (Any): value you want to ensure type for
        _types (type, tuple(type)): the data types you want to validate against. \
            E.g (str,int,float)
        val_name (str, optional): the actual name of the variable you are validating. \
            Example `ensure_type(var_1, (int, float), 'var_1'). Defaults to None.

    Raises:
        TypeError: if value does not match the validation types entered
    """
    _types = tuple(listify(_types))

    for _type in _types:
        if not type(type) == type(_type):
            raise TypeError("Expects '_type' to be a valid type, got", _type)

    # validate arguments
    if val_name and not isinstance(val_name, str):
        raise TypeError("Expects 'val_name' to be a string, got", val_name)

    if not isinstance(val, _types):
        val_name_str = f" '{val_name}' to be" if val_name else ""
        types_str = [get_type_name(_type) for _type in _types]
        types_str = ", ".join(types_str)

        raise TypeError(f"Expects{val_name_str} '{types_str}', got", val)



"""
 Copyright (c) 2023 Anthony Mugendi
 
 This software is released under the MIT License.
 https://opensource.org/licenses/MIT
"""


import re
import types
from typing import Any, Union

from .types import ensure_type


# is variable a class?
def is_class(cls) -> bool:
    """Return True if the value is a class .

    Returns:
        bool: True if class
    """
    return str(type(cls)).startswith("<class") and hasattr(cls, "__weakref__")


# check if a class is initialized
def is_initialized(cls) -> bool:
    """Returns True if the given class is initialize .

    Raises:
        TypeError: if cls argument is not a class

    Returns:
        bool: True if class has been initialized
    """

    if not is_class(cls):
        raise TypeError("Expects a valid class, got", cls)

    new_cls = types.new_class("_")
    return is_class(cls) and type(cls) != type(new_cls)


# get properties of this class
def get_props(
    cls, filter_callable: bool = False, with_meta: bool = False
) -> Union[list, dict]:
    """List all (non default) properties of a class .

    Args:
        cls (Any): a valid class
        filter_callable (bool, optional): if True, only callable properties of\
            the class are returned. Defaults to False.
        with_meta (bool, optional): if True, a dict of properties and metadata\
        about them is returned. Defaults to False.

    Returns:
        list|dict: list or dict of properties of the class
    """

    # validate inputs
    if not is_class:
        raise TypeError("Expects 'cls' to be a class, got", cls)

    ensure_type(filter_callable, bool, "filter_callable")
    ensure_type(with_meta, bool, "with_meta")

    default_props = dir(types.new_class("_"))

    resp = {} if with_meta else set()

    for p in dir(cls):
        if p not in default_props and (
            getattr(cls, p) if not filter_callable else callable(getattr(cls, p))
        ):
            if with_meta:
                resp[p] = {
                    "type": re.match(r"<class\s+'([^>']+)", str(type(getattr(cls, p))))[
                        1
                    ],
                    "value": (
                        getattr(cls, p) if not callable(getattr(cls, p)) else "N/A"
                    ),
                }

            else:
                resp.add(p)

    return resp


def get_prop_values(
    obj: Any,
    decode_bytes: bool = True,
    run_callable: bool = False,
    skip_callable: bool = True,
) -> dict:
    """Get property values of a class/object .

    Args:
        obj (Any): the object to get properties from
        decode_bytes (bool, optional): try and decode any byte values\
              to utf-8. Defaults to True.
        run_callable (bool, optional): try and run callable properties.\
              Defaults to False.
        skip_callable (bool, optional): determines if to include callable\
              values in response. Defaults to True.

    Returns:
        dict: a dict of {property:value} pairs
    """

    # validate inputs
    ensure_type(decode_bytes, bool, "decode_bytes")
    ensure_type(run_callable, bool, "run_callable")
    ensure_type(skip_callable, bool, "skip_callable")

    prop_vals = {}
    for p in dir(obj):
        val = getattr(obj, p)

        if decode_bytes and isinstance(val, bytes):
            try:
                val = val.decode()
            except UnicodeError:
                pass

        if callable(val):
            if run_callable:
                try:
                    val = val()
                except Exception as e:
                    val = e.args[0]
                    pass

            if not skip_callable:
                prop_vals[p] = val

        else:
            prop_vals[p] = val

    return prop_vals
